#include <cmath>
#include <euler_scheme_RNM.hpp>
#include "MatriceCreuse_tpl.hpp"

#if 0
#define SQUARE(x) pow((x),2)
#else
#define SQUARE(x) ((x)*(x))
#endif

#define CHGT_S2Y(s)   ((s)/(1.+(s)))
#define CHGT_Y2S(y)   ((y)/(1.-(y)))

void Euler_Scheme::energy( vector< KN<double> >& error, double & energy)
{
  int i,n;
  double dt,t,S_l,S_r,h_n;
  double e;

  for (unsigned int it=1;it<grid_t.size();it++)
    {
      n=S_steps[it].size();
      t=grid_t[it];
      dt=t-grid_t[it-1];
      if (it<grid_t.size()-1)
	dt+=grid_t[it+1] -t;
      e=0.5*dt;
//      h_n=S_steps[it][0];
      h_n=CHGT_Y2S(S_nodes[it][1])-CHGT_Y2S(S_nodes[it][0]);
//	S_steps[0][i].push_back(grid_S[0][i][j+1]-grid_S[0][i][j]);

      energy+= e*h_n/3*SQUARE( error[it](1)- error[it](0) );
      cout << "adding energy (1): " << e*h_n/3*SQUARE(  error[ it ]( 1 )- error[ it ]( 0 ) ) << endl;
      S_l=0;// inutile ?
      S_r=h_n;
      for(i=1;i< n-1;i++)
	{
	  S_l=S_r;
	  S_r=CHGT_Y2S( S_nodes[it][i+1] );
//	  h_n=S_steps[it][i];
          h_n=CHGT_Y2S(S_nodes[it][i+1])-CHGT_Y2S(S_nodes[it][i]);
	  energy+= e/3 *(S_l*(S_l+S_r) + S_r*S_r)/h_n* SQUARE( error[it](i+1)-error[it](i) );
	  cout << "hn = "<< h_n<< endl;
	  cout << "error = "<< SQUARE(   error[  it ](  i+1 )-error[  it ](  i ) )<< endl;
      cout << "adding energy (2): " << e/3 *( S_l*( S_l+S_r ) + S_r*S_r )/h_n* SQUARE(  error[ it ]( i+1 )-error[ it ]( i ) ) << endl;

	}
      S_l=S_r;
      S_r=CHGT_Y2S( S_nodes[it][i+1] );
//      h_n=S_steps[it][i];
      h_n=CHGT_Y2S(S_nodes[it][i+1])-CHGT_Y2S(S_nodes[it][i]);
//      energy+= e/3 *(S_l*(S_l+S_r) + S_r*S_r)/h_n* SQUARE( error[it](i+1)-error[it](i));
      energy+= e/3 *(S_l*(S_l+S_r) + S_r*S_r)/h_n* SQUARE(error[it](i));

    }
}


 
static double funcfplus (double sig, double h,double y)
{
  return -SQUARE(sig)/(2.*h)*
	  (
	   y*y*y*y+
	   2.*y*y*y*(h-1.)+
	   y*y*(2.*SQUARE(h)-3.*h+1.)+
	   y*h*(SQUARE(h)-2. * h+1.)+
	   SQUARE(h)*(SQUARE(h)/5.-h/2.+1./3.)
   	  );
}


static double funcgplus (double sig,double r, double h,double y)
{
  return (r-SQUARE(sig))*(y*y/2.  + (h/3.-1./2.)*y+ h*(h-2.)/12.) ;
}




void Euler_Scheme::Time_Step(int it,  vector< KN<double> >& P,int verbose)
{
  int i,n;
  double dt,t,S,h_p,h_n,r;
  double a,b;

  n=S_steps[it].size();
  MatriceProfile<double> A(n,2); 

  t=grid_t[it];
  dt=t-grid_t[it-1];
  r=rate(t);


  h_n=S_steps[it][0];

  a=dt*( funcfplus( vol( t,0 ),h_n,0. )+funcgplus( vol( t,0 ),r,h_n,0. ) );
  A( 0,1 )= a + (1.-dt*r) *h_n/6.;
  A( 0,0 )=-a + (1.+dt*r)*h_n/3.;
  for(i=1;i< n-1;i++)
    {
      h_p=h_n;
      S=S_nodes[it][i];
      h_n=S_steps[it][i];

      a=  dt*( funcfplus( vol( t,S ),h_n,S )  + funcgplus( vol( t,S ),r,h_n,S ) )   ;
      b=  -1.* dt*( funcfplus( vol( t,S ),-h_p,S ) - funcgplus( vol( t,S ),r,-h_p,S ) )  ;// on utilise des relations entre les coefficients de la matrice par rapport aux deux fonctions f et g dfinies)
      A( i,i )=-( a+b )+(1.+dt*r)*( h_n+h_p )/3.;
      A( i,i-1 )=b+(1.-dt*r)*h_p/6.;
      A( i,i+1 )=a+(1.-dt*r)*h_n/6.;
    }

  h_p=h_n;
  S=S_nodes[it][i];
  h_n=S_steps[it][i];

  a=dt*( funcfplus( vol( t,S ),h_n,S )  + funcgplus( vol( t,S ),r,h_n,S ) )   ;
  b=-1.* dt*( funcfplus( vol( t,S ),-h_p,S ) - funcgplus( vol( t,S ),r,-h_p,S ) )  ;// on utilise des relations entre les coefficients de la matrice par rapport aux deux fonctions f et g dfinies)
  A( i,i )=-( a+b )+(1.+dt*r)*( h_n+h_p )/3.;
  A( i,i-1 )=b+(1.-dt*r)*h_p/6.;
 
  //  cout<<"build_rhs at it="<<it<<endl;
  //cout<<"size of rhs ="<<P[it].N()<<"size of previous vector ="<<P[it-1].N()<< endl;
  if (change_grid[it])
    {
      if(verbose)
	build_rhs_verbose(P[it],P[it-1],S_steps[it-1],S_nodes[it-1],S_steps[it],S_nodes[it]);
      else
	build_rhs(P[it],P[it-1],S_steps[it-1],S_nodes[it-1],S_steps[it],S_nodes[it]);
    }
  else
    build_rhs_same_grid(P[it],P[it-1],S_steps[it-1],S_nodes[it-1],S_steps[it],S_nodes[it]);

  A.LU();
  A.Solve(P[it],P[it]);
}

void    Euler_Scheme::build_rhs(KN<double> &u, const KN<double> &u_p, const  vector<double> & steps_p, const  vector<double> & nodes_p,
				const  vector<double> & steps, const  vector<double> & nodes)
{
  double h;
  double u_pl,u_pr;
  double phi_l,phi_r;
  double psi_l,psi_r;
  double x_l,x_r;
  int i=0;
  int j=0;
  u=0;
  u_pl=u_p(0);
  x_l=0;
  while (i<steps.size())
    {
      //  cout<< "i=" <<i<<" j="<<j<<endl;
      if(nodes_p[j+1]<=nodes[i+1]+eps)
	{
	  phi_l=0.0;
	  psi_l=1.0;
	  while (nodes_p[j+1]<=nodes[i+1]+eps)
	    {
	      if (j<steps_p.size()-1)
		u_pr=u_p(j+1);
	      else
		u_pr=0;
	      x_r=nodes_p[j+1];
	      h=x_r-x_l;
	      phi_r=(nodes_p[j+1]-nodes[i])/steps[i];
	      psi_r=1.-phi_r;
	      if(i<steps.size()-1)
		u(i+1)+=h*(2* (phi_l*u_pl+phi_r*u_pr) + (phi_l*u_pr+phi_r*u_pl) )/6;
	      u(i)+=h*(2* (psi_l*u_pl+psi_r*u_pr) + (psi_l*u_pr+psi_r*u_pl) )/6;
	      x_l=x_r;
	      u_pl=u_pr;
	      phi_l=phi_r;
	      psi_l=psi_r;
	      j++;
	      if (j==steps_p.size())
		break;
	    }
	  if(nodes[i+1]>nodes_p[j]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      if(j<steps_p.size()-1)
		u_pr=u_p(j)+(nodes[i+1]-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      else 
		u_pr=(1.-(nodes[i+1]-nodes_p[j])/steps_p[j])*u_p(j);

	      phi_r=1.0;
	      psi_r=0.0;
	      if(i<steps.size()-1)
		u(i+1)+=h*(2* (phi_l*u_pl+phi_r*u_pr) + (phi_l*u_pr+phi_r*u_pl) )/6;
	      u(i)+=h*(2* (psi_l*u_pl+psi_r*u_pr) + (psi_l*u_pr+psi_r*u_pl) )/6;
	      u_pl=u_pr;
	      x_l=x_r;
	    }
	  i++;
	}
      else 
	{
	  while(nodes[i+1]<=nodes_p[j+1]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      if (j<steps_p.size()-1)
		u_pr=u_p(j)+(nodes[i+1]-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      else
		u_pr=u_p(j)*(1-(nodes[i+1]-nodes_p[j])/steps_p[j]);
	      if(i<steps.size()-1)
		u(i+1)+=h*(2* u_pr +u_pl) /6;
	      u(i)+=h*(2* u_pl +u_pr) /6; 
	      x_l=x_r;
	      u_pl=u_pr;
	      i++;
	      if (i==steps.size())
		break;
	    }
	  if(nodes[i]>nodes_p[j+1]-eps)
	    j++;
	}
    }
  
}










void    Euler_Scheme::build_rhs_verbose(KN<double> &u, const KN<double> &u_p, const  vector<double> & steps_p, const  vector<double> & nodes_p,
				const  vector<double> & steps, const  vector<double> & nodes)
{
  double h;
  double u_pl,u_pr;
  double phi_l,phi_r;
  double psi_l,psi_r;
  double x_l,x_r;
  int i=0;
  int j=0;
  u=0;
  u_pl=u_p(0);
  x_l=0;
  while (i<steps.size())
    {
      //  cout<< "i=" <<i<<" j="<<j<<endl;
      if(nodes_p[j+1]<=nodes[i+1]+eps)
	{
	  phi_l=0.0;
	  psi_l=1.0;
	  while (nodes_p[j+1]<=nodes[i+1]+eps)
	    {
	      u_pr=u_p(j+1);
	      x_r=nodes_p[j+1];
	      h=x_r-x_l;
	      phi_r=(nodes_p[j+1]-nodes[i])/steps[i];
	//       if (phi_r>=1+eps) 
// 		cout << "phi_r="<<phi_r<< " i="<<i<< " j="<<j<<  endl;
	      assert(phi_r<1+eps);
	      assert(phi_r>-eps);
	      psi_r=1.-phi_r;
	      cout<<"a xl="<< x_l<<" xr="<<x_r<<" nodes[i]=" <<nodes[i]  << " nodes[i+1]=" <<nodes[i+1] <<" nodes_p[j]="<<nodes_p[j] <<" nodes_p[j+1]="<<nodes_p[j+1]<<endl; 
	      u(i+1)+=h*(2* (phi_l*u_pl+phi_r*u_pr) + (phi_l*u_pr+phi_r*u_pl) )/6;
	      u(i)+=h*(2* (psi_l*u_pl+psi_r*u_pr) + (psi_l*u_pr+psi_r*u_pl) )/6;
	      cout<<"b  u("<<i<<")="<< u(i)<<" u("<<i+1<<")="<< u(i+1)<<endl; 
	      x_l=x_r;
	      u_pl=u_pr;
	      phi_l=phi_r;
	      psi_l=psi_r;
	      j++;
	      //	      cout<<"i="<<i<<" j="<<j<< " nodes[i]=" <<nodes[i]  << " nodes[i+1]=" <<nodes[i+1] <<" nodes_p[j]="<<nodes_p[j] <<" nodes_p[j+1]="<<nodes_p[j+1]<<endl;
	      if (j==steps_p.size())
		break;
	    }
	  //	  if(nodes_p[j+1]>nodes[i+1]+eps)
	    if(nodes[i+1]>nodes_p[j]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      cout<<"a xl="<< x_l<<" xr="<<x_r<<" nodes[i]=" <<nodes[i]  << " nodes[i+1]=" <<nodes[i+1] <<" nodes_p[j]="<<nodes_p[j] <<" nodes_p[j+1]="<<nodes_p[j+1]<<endl; 
	      u_pr=u_p(j)+(nodes[i+1]-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      phi_r=1.0;
	      psi_r=0.0;
	      u(i+1)+=h*(2* (phi_l*u_pl+phi_r*u_pr) + (phi_l*u_pr+phi_r*u_pl) )/6;
	      u(i)+=h*(2* (psi_l*u_pl+psi_r*u_pr) + (psi_l*u_pr+psi_r*u_pl) )/6;
	      cout<<"b  u("<<i<<")="<< u(i)<<" u("<<i+1<<")="<< u(i+1)<<endl; 
	      u_pl=u_pr;
	      x_l=x_r;
	    }
	  i++;
	}
      else 
	{
	  while(nodes[i+1]<=nodes_p[j+1]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      u_pr=u_p(j)+(nodes[i+1]-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      cout<<"c xl="<< x_l<<" xr="<<x_r<<" nodes[i]=" <<nodes[i]  << " nodes[i+1]=" <<nodes[i+1] <<" nodes_p[j]="<<nodes_p[j] <<" nodes_p[j+1]="<<nodes_p[j+1]<<endl;
	      u(i+1)+=h*(2* u_pr +u_pl) /6;
	      u(i)+=h*(2* u_pl +u_pr) /6; 
	      cout<<"d  u("<<i<<")="<< u(i)<<" u("<<i+1<<")="<< u(i+1)<<endl; 
	      x_l=x_r;
	      u_pl=u_pr;
	      i++;
	      //	      cout<<"i="<<i<<" j="<<j<< " nodes[i]=" <<nodes[i]  << " nodes[i+1]=" <<nodes[i+1] <<" nodes_p[j]="<<nodes_p[j] <<" nodes_p[j+1]="<<nodes_p[j+1]<<endl;
	      if (i==steps.size())
		break;
	    }
	  
	}
    }
  
}





void    Euler_Scheme::build_rhs_same_grid(KN<double> &u, const KN<double> &u_p, const  vector<double> & steps_p, const  vector<double> & nodes_p,
				const  vector<double> & steps, const  vector<double> & nodes)
{
  int i=0;
  u=0;
  for ( i=0; i<steps.size()-1;i++)
    {
      u(i)+=steps[i]/6.*(2*u_p(i)+u_p(i+1));
      u(i+1)+=steps[i]/6.*(u_p(i)+2*u_p(i+1));
    }
  u(i)+=steps[i]/3.*u_p(i);
}





double Euler_Scheme::build_time_error_estimator(const KN<double> &u, const KN<double> &u_p, const  vector<double> & steps_p, const  vector<double> & nodes_p,
				const  vector<double> & steps, const  vector<double> & nodes)
{
  double h;
  int i=0;
  int j=0;
  double u_pr,u_r,e_r;
  double x_l,x_r;
  double u_pl=u_p(0);
  double e_l=u(0)-u_pl;
  double estim=0;
  x_l=0;
  while (i<steps.size())
    {
      if(nodes_p[j+1]<=nodes[i+1]+eps)
	{
	  while (nodes_p[j+1]<=nodes[i+1]+eps)
	    {
	      if (j<steps_p.size()-1)
		u_pr=u_p(j+1);
	      else
		u_pr=0;
	      
	      x_r=nodes_p[j+1];
	      h=x_r-x_l;
	      // cout<<"a: h="<<h<<endl;
	      //assert (h>1e-10);
	      if(i<steps.size()-1)
		u_r=u(i)+(x_r-nodes[i])/steps[i]*(u(i+1)-u(i));
	      else 
		u_r=u(i)-(x_r-nodes[i])/steps[i]*u(i);
	      e_r=u_r-u_pr;
	      estim+= (x_r*x_r*e_r-x_l*x_l*e_l)*(e_r-e_l)/h+h*((2*e_r+e_l)*e_r+(2*e_l+e_r)*e_l)/6;
	      e_l=e_r;
	      u_pl=u_pr;
	      x_l=x_r;
	      j++;
	      if (j==steps_p.size())
		break;
	    }
	  if(nodes[i+1]>nodes_p[j]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      // cout<<"b: h="<<h<<endl;
	      //assert (h>1e-10);
	      if (j<steps_p.size()-1)
	      u_pr=u_p(j)+(x_r-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      else
		 u_pr=u_p(j)*(1.-(x_r-nodes_p[j])/steps_p[j]);
	      if (i<steps.size()-1)
		e_r=u(i+1)-u_pr;
	      else 
		e_r=-u_pr;
	      estim+= (x_r*x_r*e_r-x_l*x_l*e_l)*(e_r-e_l)/h+h*((2*e_r+e_l)*e_r+(2*e_l+e_r)*e_l)/6;
	      e_l=e_r;
	      u_pl=u_pr;
	      x_l=x_r;
	    }
	  i++;
	}
      else 
	{
	  while(nodes[i+1]<=nodes_p[j+1]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      // cout<<"c: h="<<h<<endl;
	      //assert (h>1e-10);
	      if (j<steps_p.size()-1)
		u_pr=u_p(j)+(x_r-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      else
		u_pr=u_p(j)-(x_r-nodes_p[j])/steps_p[j]*u_p(j);
	      if (i<steps.size()-1)
		e_r=u(i+1)-u_pr;
	      else
		e_r=-u_pr;
	      estim+= (x_r*x_r*e_r-x_l*x_l*e_l)*(e_r-e_l)/h+h*((2*e_r+e_l)*e_r+(2*e_l+e_r)*e_l)/6;
	      e_l=e_r;
	      u_pl=u_pr;
	      x_l=x_r;
	      i++;
	      if (i==steps.size())
		break;
	    }
	  if(nodes[i]>nodes_p[j+1]-eps)
	    j++;
	}
    }
    return estim;
}










double Euler_Scheme::build_time_error_estimator_same_grid(const KN<double> &u, const KN<double> &u_p, const  vector<double> & steps_p, const  vector<double> & nodes_p,
				const  vector<double> & steps, const  vector<double> & nodes)
{
  double estim=0;
  double e_r;
  double e_l=u(0)-u_p(0);
  for (int i=0; i<steps.size()-1;i++)
    {
      e_r=u(i+1)-u_p(i+1);
      estim+=  (nodes[i+1]*nodes[i+1]*e_r- nodes[i]*nodes[i]*e_l)*(e_r-e_l)/steps[i]+steps_p[i]*((2*e_r+e_l)*e_r+(2*e_l+e_r)*e_l)/6;
      e_l=e_r;
    }
  return estim;
}






double Euler_Scheme::build_time_error_estimator(int it,  const vector< KN<double> >& P)
{
  double dt=grid_t[it]-grid_t[it-1];
  if (change_grid[it])
    return dt*build_time_error_estimator(P[it],P[it-1],S_steps[it-1],S_nodes[it-1],S_steps[it],S_nodes[it]);
  else
    return dt*build_time_error_estimator_same_grid(P[it],P[it-1],S_steps[it-1],S_nodes[it-1],S_steps[it],S_nodes[it]);
}








void    Euler_Scheme::build_S_estimator_1(const KN<double> &u, const KN<double> &u_p, const  vector<double> & steps_p, const  vector<double> & nodes_p,
				const  vector<double> & steps, const  vector<double> & nodes, const double t , const double dt, KN<double> &estim )
{
  double x_l,x_r;
  x_l=0;
  double h;
  double u_pl,u_pr;
  double u_l,u_r;
  double v_l,v_r;
  int i=0;
  int j=0;
  double rt=rate(t);
  estim=0;
  u_pl=u_p(0);
  u_l=u(0);
  v_l=(u_l-u_pl)/dt;
  while (i<steps.size())
    {
      //  cout<< "i=" <<i<<" j="<<j<<endl;
      if(nodes_p[j+1]<=nodes[i+1]+eps)
	{
	  while (nodes_p[j+1]<=nodes[i+1]+eps)
	    {
	      x_r=nodes_p[j+1];
	      h=x_r-x_l;
	      if (j<steps_p.size()-1)
		u_pr=u_p(j+1);
	      else
		u_pr=0;
	      if(i<steps.size()-1)
		u_r=u(i)+(x_r-nodes[i])/steps[i]*(u(i+1)-u(i));
	      else
		u_r=u(i)-(x_r-nodes[i])/steps[i]*u(i);
	      v_r=(u_r-u_pr)/dt;
	      estim(i)-=rt* (u_r-u_l)/3* (x_l*(2*v_l+v_r)+x_r*(2*v_r+v_l));
	      estim(i)+= ((2*rt*u_l+v_l)*(2*v_l+v_r)+(2*rt*u_r+v_r)*(2*v_r+v_l))*h/6;
	      x_l=x_r;
	      u_pl=u_pr;
	      u_l=u_r;
	      v_l=v_r;
	      j++;
	      if (j==steps_p.size())
		break;
	    }
	  if(nodes[i+1]>nodes_p[j]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      if (j<steps_p.size()-1)
		u_pr=u_p(j)+(nodes[i+1]-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      else
		u_pr=u_p(j)*(1.-(nodes[i+1]-nodes_p[j])/steps_p[j]);
	      if(i<steps.size()-1)
		u_r=u(i+1);
	      else
		u_r=0; 
	      v_r=(u_r-u_pr)/dt;
	      estim(i)-=rt* (u_r-u_l)/3* (x_l*(2*v_l+v_r)+x_r*(2*v_r+v_l));
	      estim(i)+= ((2*rt*u_l+v_l)*(2*v_l+v_r)+(2*rt*u_r+v_r)*(2*v_r+v_l))*h/6;
	      u_pl=u_pr;
	      x_l=x_r;
	      u_l=u_r;
	      v_l=v_r;
	    }
	  i++;
	}
      else 
	{
	  while(nodes[i+1]<=nodes_p[j+1]+eps)
	    {
	      x_r=nodes[i+1];
	      h=x_r-x_l;
	      if (j<steps_p.size()-1)
		u_pr=u_p(j)+(x_r-nodes_p[j])/steps_p[j]*(u_p(j+1)-u_p(j));
	      else
		u_pr=u_p(j)-(x_r-nodes_p[j])/steps_p[j]*u_p(j);
	      if(i<steps.size()-1)
		u_r=u(i+1);
	      else
		u_r=0;
	      v_r==(u_r-u_pr)/dt;
	      estim(i)-=rt* (u_r-u_l)/3* (x_l*(2*v_l+v_r)+x_r*(2*v_r+v_l));
	      estim(i)+=((2*rt*u_l+v_l)*(2*v_l+v_r)+(2*rt*u_r+v_r)*(2*v_r+v_l))*h/6;
	      u_pl=u_pr;
	      u_l=u_r;
	      v_l=v_r;
	      x_l=x_r;
	      i++;
	      if (i==steps.size())
		break;
	    }
	  if(nodes[i]>nodes_p[j+1]-eps)
	    j++;
	}
    }
  //  cout<< "while loop done in build_S_estimator"<<endl;
   
  for(i=0;i<steps.size()-1;i++)
    {
      estim(i)+=SQUARE( rt* (u(i+1)-u(i))) *(nodes[i+1]* (nodes[i+1]+nodes[i])+nodes[i]*nodes[i])/3/steps[i];
      estim(i)-=SQUARE( rt)*(nodes[i+1]*u(i+1)*u(i+1)-nodes[i]*u(i)*u(i));
      estim(i)+=SQUARE( rt)/3*steps[i]*(u(i+1)*(2*u(i+1)+u(i))+u(i)*(2*u(i)+u(i+1) ));
    }
   estim(i)+=SQUARE( rt* (-u(i))) *(nodes[i+1]* (nodes[i+1]+nodes[i])+nodes[i]*nodes[i])/3/steps[i];
   estim(i)-=SQUARE( rt)*(-nodes[i]*u(i)*u(i));
   estim(i)+=SQUARE( rt)/3*steps[i]*(u(i)*2*u(i));

   //  cout<< "for loop 1 done in build_S_estimator"<<endl;
   for(i=0;i<steps.size();i++)
    {
      estim(i)=sqrt(fabs(estim(i)));
      estim(i)*=steps[i]/nodes[i+1];
    }
   //  cout<< "for loop 2 done in build_S_estimator"<<endl;
  for(i=0;i<steps.size()-2;i++)
    {
      double a=nodes[i+1]*SQUARE(vol(t,nodes[i+1]));
      double b=a/steps[i]* (u(i+1)-u(i));
      b+=a/steps[i+1]* (u(i+1)-u(i+2));
      b=0.25*fabs(b);
     if(i<steps.size()-1)
       {
	 estim(i)+=sqrt(steps[i])*b;
       	 estim(i+1)+=sqrt(steps[i+1])*b;
       }
    }
  double a=nodes[i+1]*SQUARE(vol(t,nodes[i+1]));
  double b=a/steps[i]* (u(i+1)-u(i));
  b+=a/steps[i+1]* (u(i+1));
  b=0.25*fabs(b);
  if(i<steps.size()-1)
    {
      estim(i)+=sqrt(steps[i])*b;
      estim(i+1)+=sqrt(steps[i+1])*b;
    }

  //  cout<< "for loop 3 done in build_S_estimator"<<endl;
}



void Euler_Scheme::build_S_estimator_1(int it,  const vector< KN<double> >& P, vector< KN<double> >& estim)
{
  double dt=grid_t[it]-grid_t[it-1];
  double r=rate(grid_t[it]);
  if (change_grid[it])
    build_S_estimator_1(P[it],P[it-1],S_steps[it-1],S_nodes[it-1],S_steps[it],S_nodes[it],grid_t[it],dt,estim[it]);
  else
    build_S_estimator_1(P[it],P[it-1],S_steps[it-1],S_nodes[it-1],S_steps[it],S_nodes[it],grid_t[it],dt,estim[it]);
}


double Euler_Scheme::build_global_S_estimator(int it, const KN<double> & estim,const  vector<double> & nodes)
{
  double dt=grid_t[it]-grid_t[it-1];
  double sum=0;
  double sigma_min=10000;
  double vola;
  double aux;
  for (int i=0;i<estim.size();i++)
    {
      aux=estim(i);
      sum+=aux*aux;
      vola=vol(grid_t[it],nodes[i+1]);
      if (vola<sigma_min)
	sigma_min=vola;
    }
  return sqrt(dt*sum)/sigma_min;
}


